/* Copyright 2009 Jetro Lauha - http://iki.fi/jetro/
 *
 * $Id: MainState.cpp 68 2009-08-30 22:13:41Z tonic $
 * $Revision: 68 $
 */

#include <assert.h>
#include <algorithm>

#include "main.h"
#include "MainState.h"


#define MAX(a, b) ((a) > (b) ? (a) : (b))
#define MIN(a, b) ((a) < (b) ? (a) : (b))


#define GRAVITY     0.0005f
#define DAMPING     0.995f

#define PLAYER_R        0.9f
#define PLAYER_R_VIS    1.0f

#define BOUNCINESS  0.5f
#define FRICTION    0.95f
#define THRUST      0.002f
#define SPEED_THRESHOLD     0.1f    /* dig vs collide */

#define BORDER_CASH_PENALTY 10

#define DIG_SOUND_FRAMES    5
#define MUSIC_VOL           255
#define SFX_VOL             160


static float sMaxRockRadius = 0.0f; // loadLevel should fill this


enum SAMPLE_ID
{
    SAMPLE_START = 0,
    SAMPLE_THRUST,
    SAMPLE_DIG,
    SAMPLE_PICKUP,
    SAMPLE_HIT,
    SAMPLE_BORDER_EXPL,
    SAMPLE_COUNT //
};

static char *sSampleFileNames[SAMPLE_COUNT] =
{
    "data/start.ogg",
    "data/thrust-loop.wav",
    "data/dig-loop.wav",
    "data/pickup.ogg",
    "data/hit.ogg",
    "data/expl1.ogg"
};

#define AUDIO_CHANNELS 32

static FMUSIC_MODULE *sMusic;
static FSOUND_SAMPLE *sSamples[SAMPLE_COUNT];


MainState::MainState()
{
    for (int a = 0; a < KID_COUNT; ++a)
        mKey[a] = false;
    mThrustChannel = -1;
    mDigChannel = -1;
    mDigCounter = 0;
    mEndingUIActive = false;
}


MainState::~MainState()
{
}


bool MainState::init()
{
    Graphics *g = App::getInstance()->getGraphics();

    mTitleUIActive = true;
    mInGame = false;

    if (!FSOUND_Init(44100, AUDIO_CHANNELS, 0))
    {
        gErrorMessage("Audio error", "Unable to initialize sound! (%d)",
                      FSOUND_GetError());
    }
    for (u32 a = 0; a < SAMPLE_COUNT; ++a)
    {
        sSamples[a] = FSOUND_Sample_Load(FSOUND_FREE, sSampleFileNames[a], FSOUND_LOOP_OFF, 0, 0);
        if (!sSamples[a])
            gErrorMessage("Audio error", "Can't load %s", sSampleFileNames[a]);
    }
    sMusic = FMUSIC_LoadSong("data/ld48-15.it");
    if (!sMusic)
        gErrorMessage("Audio error", "Can't load music");
    FMUSIC_PlaySong(sMusic);
    FMUSIC_SetMasterVolume(sMusic, MUSIC_VOL);

    mFont = new BMFont(g, "data/droid30.fnt");
    if (!mFont->init())
        return false;

    // Make sure these are in synch with IMAGE_ID
    mImages.push_back(g->loadImage("data/button.png"));
    mImages.push_back(g->loadImage("data/button-hot.png"));
    mImages.push_back(g->loadImage("data/button-active.png"));
    mImages.push_back(g->loadImage("data/button-hot-active.png"));
    mImages.push_back(g->loadImage("data/cave-circle.png", IF_COLOR | IF_ALPHA | IF_ALIGN_CENTER | IF_FILTER_TRILINEAR));
    mImages.push_back(g->loadImage("data/ship.png", IF_COLOR | IF_ALPHA | IF_ALIGN_CENTER | IF_FILTER_TRILINEAR));
    mImages.push_back(g->loadImage("data/ship-borderline.png", IF_COLOR | IF_ALPHA | IF_ALIGN_CENTER | IF_FILTER_TRILINEAR));
    mImages.push_back(g->loadImage("data/flare.png", IF_COLOR | IF_ALPHA | IF_ALIGN_CENTER | IF_FILTER_TRILINEAR));
    mImages.push_back(g->loadImage("data/flame.png", IF_COLOR | IF_ALPHA | IF_ALIGN_CENTER | IF_FILTER_TRILINEAR));

    mButtonStyle = new IMGUIImageButtonStyle(mFont, mImages[IMG_BUTTON], mImages[IMG_BUTTON_HOT], mImages[IMG_BUTTON_ACTIVE], mImages[IMG_BUTTON_HOT_ACTIVE]);

    mLabelText = "";

    mDevice = App::getInstance()->getIrrDevice();
    mVideo = App::getInstance()->getIrrVideoDriver();
    mSmgr = App::getInstance()->getIrrSceneManager();


    //mCamera = mSmgr->addCameraSceneNodeMaya(0, -50, 10, 30);
    mCamera = mSmgr->addCameraSceneNode();
    mCamera->setNearValue(0.1f);
    mCamera->setFarValue(1200);
    mCamera->setPosition(core::vector3df(0, 0, 0));
    mCamera->setTarget(core::vector3df(0, 0, 1));
    mCamera->setUpVector(core::vector3df(0, 1, 0));
    mSmgr->setActiveCamera(mCamera);

    video::ITexture *skyDomeTex = mVideo->getTexture("data/sky.png");
    scene::ISceneNode *skyDomeNode = mSmgr->addSkyDomeSceneNode(skyDomeTex, 16, 8, 1, 2);
    //skyDomeNode->getMaterial(0).setFlag(video::EMF_BILINEAR_FILTER, false);
    skyDomeNode->getMaterial(0).getTextureMatrix(0).setScale(core::vector3df(4, 2, 1));

    mSmgr->setShadowColor(SColor(0x14001040));

    mLastMousePos.set(WIDTH / 2, HEIGHT / 2);

    loadLevel();

    mCash = 200.f;

    return mFont != NULL && mButtonStyle != NULL;
}


void MainState::deinit()
{
    UI32 a;
    delete mFont;
    for (a = 0; a < mImages.size(); ++a)
        delete mImages[a];
    mImages.clear();

    FMUSIC_StopAllSongs();
    FMUSIC_FreeSong(sMusic);
    for (a = 0; a < AUDIO_CHANNELS; ++a)
        FSOUND_StopSound(a);
    for (a = 0; a < SAMPLE_COUNT; ++a)
        FSOUND_Sample_Free(sSamples[a]);
    FSOUND_Close();
}


void MainState::loadLevel()
{
    mLevelCells.clear();
    IImage *img = App::getInstance()->getIrrVideoDriver()->createImageFromFile("data/level.png");
    assert(img);
    mLevelWidth = (int)img->getDimension().Width;
    mLevelHeight = (int)img->getDimension().Height;
    srand(123);
    mTreasureCount = 0;
    for (int y = 0; y < mLevelHeight; ++y)
    {
        for (int x = 0; x < mLevelWidth; ++x)
        {
            LevelCell lc;
            lc.color = img->getPixel((u32)x, (u32)y);
            lc.x = (float)x;
            lc.y = (float)y;
            lc.x += ((rand() & 0x7fff) / 32767.f) * 2 / 3 - 1.f / 3;
            lc.y += ((rand() & 0x7fff) / 32767.f) * 2 / 3 - 1.f / 3;
            lc.r = 1.0f + ((rand() & 0x7fff) / 32767.f) * 0.2f;
            sMaxRockRadius = 1.2f;
            lc.flags = 0;
            if ((lc.color.color & 0xffffff) == 0)
                lc.flags |= LevelCell::LCF_EMPTY | LevelCell::LCF_NO_COLLIDER;
            if (lc.color.color == 0xff00ffff || lc.color.color == 0xffff00ff)
            {
                if (lc.color.color == 0xffff00ff)
                    mPlayerPos.set((float)x, (float)y);
                lc.color.set(255, 235 - y * 4, 240 - y * 2, 255 - y);
                lc.flags |= LevelCell::LCF_NO_COLLIDER;
            }
            if (lc.color.getBlue() == 0 && lc.color.getRed() > 190)
            {
                lc.flags |= LevelCell::LCF_MOVING;
                if (lc.color.color == 0xffffff00)
                {
                    lc.flags &= ~LevelCell::LCF_MOVING;
                    lc.flags |= LevelCell::LCF_GOLD;
                    ++mTreasureCount;
                }
            }
            mLevelCells.push_back(lc);
        }
    }
    img->drop();
}


void MainState::onActivate(UI32 time)
{
    AppState::onActivate(time);
    mStartTime = 0; // first call to update will do actual initialization
    mPaused = false;
}


bool MainState::onKeyUp(const irr::SEvent::SKeyInput &event)
{
    if (event.Key == KEY_LEFT)
        mKey[KID_LEFT] = false;
    if (event.Key == KEY_RIGHT)
        mKey[KID_RIGHT] = false;
    if (event.Key == KEY_UP)
        mKey[KID_UP] = false;
    if (event.Key == KEY_DOWN)
        mKey[KID_DOWN] = false;
    return false;
}

bool MainState::onKeyDown(const irr::SEvent::SKeyInput &event)
{
    if (event.Key == KEY_ESCAPE)
        App::getInstance()->setAlive(false);

    if (event.Key == KEY_LEFT)
        mKey[KID_LEFT] = true;
    if (event.Key == KEY_RIGHT)
        mKey[KID_RIGHT] = true;
    if (event.Key == KEY_UP)
        mKey[KID_UP] = true;
    if (event.Key == KEY_DOWN)
        mKey[KID_DOWN] = true;

    return false;
}

bool MainState::onMouseDown(const irr::SEvent::SMouseInput & /*event*/)
{
    if (mTitleUIActive || mEndingUIActive)
        mIMGUI.onMouseDown();

    return false;
}

bool MainState::onMouseUp(const irr::SEvent::SMouseInput & /*event*/)
{
    mIMGUI.onMouseUp();
    return false;
}

bool MainState::onMouseMotion(const irr::SEvent::SMouseInput &event)
{
    mLastMousePos.set(event.X, event.Y);
    mIMGUI.onMouseMotion(event.X, event.Y);
    return false;
}

bool MainState::onMouseWheel(const irr::SEvent::SMouseInput & /*event*/)
{
    return false;
}


bool MainState::update(UI32 time)
{
    bool result = AppState::update(time);
    if (!result)
        return false; // game is paused...

    if (mTitleUIActive || mEndingUIActive)
    {
        return false;
    }

    mPlayerPos += mPlayerSpeed;
    mPlayerSpeed *= DAMPING;
    mPlayerSpeed.Y += GRAVITY;

    bool thrustSound = mKey[KID_LEFT] || mKey[KID_RIGHT] || mKey[KID_UP] || mKey[KID_DOWN];
    if (thrustSound && mThrustChannel < 0)
    {
        mThrustChannel = FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_THRUST]);
        FSOUND_SetVolume(mThrustChannel, SFX_VOL);
        FSOUND_SetLoopMode(mThrustChannel, FSOUND_LOOP_NORMAL);
    }
    else if (!thrustSound && mThrustChannel >= 0)
    {
        FSOUND_StopSound(mThrustChannel);
        mThrustChannel = -1;
    }

    if (mDigCounter > 0 && mDigChannel < 0)
    {
        mDigChannel = FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_DIG]);
        FSOUND_SetVolume(mDigChannel, SFX_VOL);
        FSOUND_SetLoopMode(mDigChannel, FSOUND_LOOP_NORMAL);
    }
    else if (mDigCounter <= 0 && mDigChannel >= 0)
    {
        FSOUND_StopSound(mDigChannel);
        mDigChannel = -1;
    }
    else if (mDigCounter > 0 && mDigChannel >= 0)
    {
        FSOUND_SetVolume(mDigChannel, mDigCounter * SFX_VOL / DIG_SOUND_FRAMES);
    }
    if (mDigCounter > 0)
        --mDigCounter;

    if (mKey[KID_LEFT])
        mPlayerSpeed.X -= THRUST;
    if (mKey[KID_RIGHT])
        mPlayerSpeed.X += THRUST;
    if (mKey[KID_UP])
        mPlayerSpeed.Y -= THRUST;
    if (mKey[KID_DOWN])
        mPlayerSpeed.Y += THRUST;

    // player boundingbox top-left, bottom-right:
    vector2df plrBBTL = mPlayerPos - vector2df(PLAYER_R, PLAYER_R);
    vector2df plrBBBR = mPlayerPos + vector2df(PLAYER_R, PLAYER_R);
    int gLeft = (int)floor(plrBBTL.X - sMaxRockRadius);
    int gRight = (int)floor(plrBBBR.X + sMaxRockRadius);
    int gTop = (int)ceil(plrBBTL.Y - sMaxRockRadius);
    int gBottom = (int)ceil(plrBBBR.Y + sMaxRockRadius);
    if (gLeft < 0) gLeft = 0;
    if (gRight >= mLevelWidth) gRight = mLevelWidth - 1;
    if (gTop < 0) gTop = 0;
    if (gBottom >= mLevelHeight) gBottom = mLevelHeight - 1;
    for (int y = gTop; y <= gBottom; ++y)
    {
        for (int x = gLeft; x <= gRight; ++x)
        {
            LevelCell &lc = mLevelCells[y * mLevelWidth + x];
            if (lc.flags & LevelCell::LCF_NO_COLLIDER)
                continue;
            float comboR = lc.r + PLAYER_R;
            vector2df rockPos(lc.x, lc.y);
            if (mPlayerPos.getDistanceFromSQ(rockPos) < comboR * comboR)
            {
                // first project player out of collision
                float distance = mPlayerPos.getDistanceFrom(rockPos);
                float penetrationLength = comboR - distance;
                vector2df normal = mPlayerPos - rockPos;
                normal.normalize();
                vector2df penetration = penetrationLength * -normal;
                mPlayerPos -= penetration;

                float playerSpeedSq = mPlayerSpeed.getLengthSQ();
                if (playerSpeedSq < SPEED_THRESHOLD * SPEED_THRESHOLD)
                {
                    // speed below threshold: dig rock
                    mDigCounter = DIG_SOUND_FRAMES;
                    lc.r -= penetrationLength;
                    mCash -= penetrationLength;
                    if (lc.r <= 0)
                    {
                        lc.flags |= LevelCell::LCF_EMPTY | LevelCell::LCF_NO_COLLIDER;
                        if (lc.flags & LevelCell::LCF_GOLD)
                        {
                            int chn = FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_PICKUP]);
                            FSOUND_SetVolume(chn, SFX_VOL);
                            mCash += 100;
                            --mTreasureCount;
                        }
                        else if (lc.flags & LevelCell::LCF_MOVING)
                        {
                            int chn = FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_BORDER_EXPL]);
                            FSOUND_SetVolume(chn, SFX_VOL);
                            mCash -= 50;
                        }
                    }
                }
                else
                {
                    // speed above threshold: reflect the player speed
                    vector2df perpendicular = -normal * mPlayerSpeed.dotProduct(-normal);
                    vector2df parallel = mPlayerSpeed - perpendicular;
                    mPlayerSpeed = parallel * FRICTION - perpendicular * BOUNCINESS;
                    int chn = FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_HIT]);
                    FSOUND_SetVolume(chn, SFX_VOL);
                }
            }
        }
    }

    if (mPlayerPos.X < 0)
    {
        mPlayerPos.X = 0;
        if (mPlayerSpeed.X < 0)
            mPlayerSpeed.X = -mPlayerSpeed.X;
        mCash -= BORDER_CASH_PENALTY;
        int chn = FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_BORDER_EXPL]);
        FSOUND_SetVolume(chn, SFX_VOL);
    }
    if (mPlayerPos.X > (float)mLevelWidth)
    {
        mPlayerPos.X = (float)mLevelWidth;
        if (mPlayerSpeed.X > 0)
            mPlayerSpeed.X = -mPlayerSpeed.X;
        mCash -= BORDER_CASH_PENALTY;
        int chn = FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_BORDER_EXPL]);
        FSOUND_SetVolume(chn, SFX_VOL);
    }
    if (mPlayerPos.Y < 0)
    {
        mPlayerPos.Y = 0;
        if (mPlayerSpeed.Y < 0)
            mPlayerSpeed.Y = -mPlayerSpeed.Y;
        mCash -= BORDER_CASH_PENALTY;
        int chn = FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_BORDER_EXPL]);
        FSOUND_SetVolume(chn, SFX_VOL);
    }
    if (mPlayerPos.Y > (float)mLevelHeight)
    {
        mPlayerPos.Y = (float)mLevelHeight;
        if (mPlayerSpeed.Y > 0)
            mPlayerSpeed.Y = -mPlayerSpeed.Y;
        mCash -= BORDER_CASH_PENALTY;
        int chn = FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_BORDER_EXPL]);
        FSOUND_SetVolume(chn, SFX_VOL);
    }


    if (mInGame && (mCash <= 0 || mTreasureCount <= 0))
    {
        mInGame = false;
        mEndingUIActive = true;
        int chn = FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_START]);
        FSOUND_SetVolume(chn, SFX_VOL);
    }

    return true;
}


void MainState::genCone(float x, float y, float r, const video::SColor &color)
{
    if (mIndices.size() >= 65536 - 3)
    {
        assert(0 && "too much triangles");
        return;
    }
    const float R_EXPAND = 128.f / (128 - 4 - 1);   // texture has 4 px empty border, 1 for half of outline
    r *= R_EXPAND;
    const float centerU = 0.5f, centerV = 0.5f;
    const float uvW = 0.5f, uvH = 0.5f;
    const float MAX_R = 128;
    const float Z_NEAR = 0, Z_FAR = 1.0f;//MIN(r / MAX_R, 1.0f);
    const int SEGMENTS = 12;
    const float SEG_RAD = (float)(2 * PI / SEGMENTS);
    const int preVtxCount = (int)mVertices.size();
    static core::position2df unitCircleSegmentVertices[SEGMENTS];
    static bool unitCircleSegmentVerticesInitialized = false;
    if (!unitCircleSegmentVerticesInitialized)
    {
        float angle = 0;
        for (int a = 0; a < SEGMENTS; ++a, angle += SEG_RAD)
            unitCircleSegmentVertices[a].set(cosf(angle), sinf(angle));
        unitCircleSegmentVerticesInitialized = true;
    }
    int idx = preVtxCount;
    S3DVertex vtx(x, y, Z_NEAR, 0, 0, Z_NEAR, color, centerU, centerV);
    mVertices.push_back(vtx);
    for (int seg = 0; seg < SEGMENTS; ++seg)
    {
        vtx.Pos.set(x + unitCircleSegmentVertices[seg].X * r, y + unitCircleSegmentVertices[seg].Y * r, Z_FAR);
        vtx.TCoords.set(unitCircleSegmentVertices[seg].X * 0.5f + 0.5f, unitCircleSegmentVertices[seg].Y * 0.5f + 0.5f);
        vtx.Normal.set(unitCircleSegmentVertices[seg].X, unitCircleSegmentVertices[seg].Y, 0);
        mVertices.push_back(vtx);
    }
    assert(mIndices.size() < 65536 - SEGMENTS * 3);
    for (int seg = 0; seg < SEGMENTS; ++seg)
    {
        mIndices.push_back((u16)idx);   // 1st vertex for each tri is the center one
        mIndices.push_back((u16)idx + 1 + seg);
        mIndices.push_back((u16)idx + 1 + (seg + 1) % SEGMENTS);
    }
}


void MainState::render(irr::IrrlichtDevice *device)
{
    assert(device);
    irr::video::IVideoDriver *driver = device->getVideoDriver();
    assert(driver);
    irr::scene::ISceneManager *smgr = device->getSceneManager();

    static core::array<vector2df> flarePositions;

    driver->beginScene(true, true, video::SColor(0x00445566));

    smgr->drawAll();

    Graphics *g = App::getInstance()->getGraphics();
    g->beginRender();

    int tmprandseed = rand();
    if (mInGame)
    {
        mCamAngle = mPlayerPos.X * -0.03f;
        mCamera->setTarget(vector3df(cosf(mCamAngle), 0, sinf(mCamAngle)));

        float scale = 27 + mPlayerPos.Y / 2.7f;
        if (scale >= 64)
            scale = 64;

        const dimension2du &screenSize = driver->getScreenSize();
        const vector2df halfScreen(screenSize.Width / 2.f, screenSize.Height / 2.f);
        vector2df scaledHalfArea(screenSize.Width / 2 / scale, screenSize.Height / 2 / scale);
        vector2df screenTL = mPlayerPos * scale - halfScreen;
        screenTL.X = MIN(mLevelWidth * scale - screenSize.Width, screenTL.X);
        screenTL.Y = MIN(mLevelHeight * scale - screenSize.Height, screenTL.Y);
        screenTL.X = MAX(0, screenTL.X);
        screenTL.Y = MAX(0, screenTL.Y);

        //vector2df areaTL = mPlayerPos - scaledHalfArea - vector2df(sMaxRockRadius, sMaxRockRadius);
        //vector2df areaBR = mPlayerPos + scaledHalfArea + vector2df(sMaxRockRadius, sMaxRockRadius);
        vector2df areaTL = screenTL / scale - vector2df(sMaxRockRadius, sMaxRockRadius);
        int gTop = (int)ceil(areaTL.Y);
        int gLeft = (int)floor(areaTL.X);
        //int gBottom = (int)ceil(areaBR.Y);
        //int gRight = (int)floor(areaBR.X);
        if (gTop < 0) gTop = 0;
        if (gLeft < 0) gLeft = 0;
        int gBottom = gTop + (int)ceil(scaledHalfArea.Y * 2 + sMaxRockRadius * 2);
        int gRight = gLeft + (int)ceil(scaledHalfArea.X * 2 + sMaxRockRadius * 2);
        if (gBottom >= mLevelHeight) gBottom = mLevelHeight - 1;
        if (gRight >= mLevelWidth) gRight = mLevelWidth - 1;

        flarePositions.set_used(0);
        mVertices.set_used(0);
        mIndices.set_used(0);

        video::SMaterial mat;
        mat.MaterialType = video::EMT_ONETEXTURE_BLEND;
        mat.MaterialTypeParam = video::pack_texureBlendFunc(video::EBF_SRC_ALPHA, video::EBF_ONE_MINUS_SRC_ALPHA);
        mat.setFlag(video::EMF_LIGHTING, false);
        mat.setTexture(0, ((ImageIrr *)mImages[IMG_CAVE_CIRCLE])->getIrrTexture());
        srand(0);
        for (int y = gTop; y <= gBottom; ++y)
        {
            for (int x = gLeft; x <= gRight; ++x)
            {
                LevelCell &lc = mLevelCells[y * mLevelWidth + x];
                if (lc.flags & LevelCell::LCF_EMPTY)
                    continue;
                //srand(y * 100 + x);
                const SColor &c = lc.color;
                float xa = -screenTL.X, ya = -screenTL.Y;
                if (lc.flags & LevelCell::LCF_MOVING)
                {
                    xa += cosf(mTime * 0.001f + x * 0.5f /*+ rand() % 100*/) * scale / 6;
                    ya += sinf(mTime * 0.001f + y * 0.5f /*+ rand() % 100*/) * scale / 6;
                }
                float xf = lc.x * scale + xa, yf = lc.y * scale + ya;
                if (lc.flags & LevelCell::LCF_GOLD)
                    flarePositions.push_back(vector2df(xf, yf));
                //if (xf < -scale || xf > (float)screenSize.Width + scale || yf < -scale || yf > (float)screenSize.Height + scale)
                //    continue;
                const float r = lc.r * scale;
                genCone(xf, yf, r, c);
            }
        }
        driver->setMaterial(mat);
        driver->drawVertexPrimitiveList(mVertices.const_pointer(), mVertices.size(),
                                        mIndices.const_pointer(), mIndices.size() / 3,
                                        video::EVT_STANDARD, scene::EPT_TRIANGLES, video::EIT_16BIT);

        float plrNormalizedScale = 2.0f / (mImages[IMG_SHIP]->getWidth() - 2 * (4 + 2));    // img has 4 pix border, 2 pix for half of outline
        mImages[IMG_SHIP]->setScale(PLAYER_R_VIS * scale * plrNormalizedScale);
        float playerSpeed = (mPlayerSpeed.getLength() - SPEED_THRESHOLD) * 100;
        if (playerSpeed < 0) playerSpeed = 0;
        if (playerSpeed > 1) playerSpeed = 1;
        //Color plrColor((UI8)(145 + playerSpeed * 15),
        //               (UI8)(200 - playerSpeed * 40),
        //               (UI8)(140 + playerSpeed * 15), 255);
        Color plrColor((UI8)180, (UI8)200, (UI8)140, (UI8)255);
        mImages[IMG_SHIP]->draw(IDM_NORMAL, IF_COLOR | IF_SCALE | IF_TRANSFORM | IF_ALIGN_CENTER, mPlayerPos.X * scale - screenTL.X, mPlayerPos.Y * scale - screenTL.Y, plrColor);
        Color shipBorderlineColor((UI8)(150 - playerSpeed * 50), 75, 70, (UI8)(255 - playerSpeed * 255));
        mImages[IMG_SHIP_BORDERLINE]->setScale(PLAYER_R_VIS * scale * plrNormalizedScale + (sinf(mTime * 0.01f) + 1) * plrNormalizedScale);
        const float plrX = mPlayerPos.X * scale - screenTL.X;
        const float plrY = mPlayerPos.Y * scale - screenTL.Y;
        mImages[IMG_SHIP_BORDERLINE]->draw(IDM_NORMAL, IF_COLOR | IF_ALPHA | IF_SCALE | IF_TRANSFORM | IF_ALIGN_CENTER, plrX, plrY, shipBorderlineColor);

        const int flareCount = flarePositions.size();
        mImages[IMG_FLARE]->setScale(scale * 2.0f / (mImages[IMG_FLARE]->getWidth()));
        for (int a = 0; a < flareCount; ++a)
        {
            const vector2df &pos = flarePositions[a];
            mImages[IMG_FLARE]->draw(pos.X, pos.Y);
        }

        static const float flameAngles[4] = { (float)PI, 0, (float)-PI / 2, (float)PI / 2 };
        static const float flameXOffsets[4] = { 1.0f, -1.0f, 0, 0 };
        static const float flameYOffsets[4] = { 0, 0, 1.0f, -1.0f };
        srand(tmprandseed);
        for (int a = 0; a < KID_COUNT; ++a)
        {
            if (!mKey[a])
                continue;
            mImages[IMG_FLAME]->setScale(0.5f * scale * 2.0f / mImages[IMG_FLAME]->getWidth());
            mImages[IMG_FLAME]->setAngle(flameAngles[a] + (((rand() >> 8) & 0x7f) / 127.f - 0.5f) * 0.1f);
            mImages[IMG_FLAME]->draw(IDM_ADD, IF_COLOR | IF_ALPHA | IF_SCALE | IF_TRANSFORM | IF_ALIGN_CENTER | IF_ROTATE,
                                     plrX + flameXOffsets[a] * scale * 1.4f, plrY + flameYOffsets[a] * scale * 1.4f);
        }
    }

    mIMGUI.begin(mTime);

    if (mTitleUIActive)
    {
        mCamAngle = mTime * 0.00005f;
        mCamera->setTarget(vector3df(cosf(mCamAngle), 0, sinf(mCamAngle)));
        mFont->drawString(WIDTH / 2, 150, "Rock Warrior", IF_ALIGN_HORIZ_CENTER, 1, Color(0xffffffff));
        mFont->drawString(WIDTH / 2, 180, "Loot the gold!     Beware lava.", IF_ALIGN_HORIZ_CENTER, 0.7f, Color(0xaaffcc00));
        mImages[IMG_FLARE]->setScale(0.5f);
        float wave = sinf(mTime * 0.0015f);
        mImages[IMG_FLARE]->draw(IDM_ADD, IF_COLOR | IF_ALIGN_CENTER, wave * 90 + WIDTH / 2, 180 + mFont->getFontHeight(0.5f), Color(0.7f, 0.35f - wave * 0.3f, 0.25f - wave * 0.25f, 1.0f));

        mFont->drawString(WIDTH / 2, 250, "Move ship with cursor keys.", IF_ALIGN_HORIZ_CENTER, 0.7f, Color(0xaaffffff));
        mFont->drawString(WIDTH / 2, 280, "Ship automatically digs when moving slowly.", IF_ALIGN_HORIZ_CENTER, 0.7f, Color(0xaaffffff));

        if (mIMGUI.button(WIDTH / 2 - mButtonStyle->getWidth() / 2, 350, IMGUI_AI_COUNTER, mButtonStyle, 0, 0, "Play"))
        {
            mTitleUIActive = false;
            mEndingUIActive = false;
            mInGame = true;
            mStartTime = 0;
            int chn = FSOUND_PlaySound(FSOUND_FREE, sSamples[SAMPLE_START]);
            FSOUND_SetVolume(chn, SFX_VOL);
        }
        if (mIMGUI.button(WIDTH / 2 - mButtonStyle->getWidth() / 2, 450, IMGUI_AI_COUNTER, mButtonStyle, 0, 0, "Quit"))
            App::getInstance()->setAlive(false);

        mFont->drawString(WIDTH / 2, HEIGHT - 5, "Copyright 2009 Jetro Lauha - www.jet.ro", IF_ALIGN_BOTTOM_HCENTER, 0.5f, Color(0x80ffffff));
    }
    else if (mInGame)
    {
        char buf[100];
        sprintf(buf, "$%.1f", mCash);
        mFont->drawString(40, 40, "Cash", IF_ALIGN_TOP_LEFT, 0.7f, Color(0x88000000));
        mFont->drawString(40, 60, buf, IF_ALIGN_TOP_LEFT, 1, Color(0xff000000));
    }
    else if (mEndingUIActive)
    {
        char *title = "Looting Successful!";
        char buf[100];
        sprintf(buf, "$%.1f", mCash);
        if ((int)mCash <= 0)
        {
            title = "Failboat!";
            mCash = 0;
        }
        mFont->drawString(WIDTH / 2, 200, title, IF_ALIGN_HORIZ_CENTER, 1, Color(0xffffffff));
        mFont->drawString(WIDTH / 2, 300, "Final Cash:", IF_ALIGN_HORIZ_CENTER, 0.6f, Color(0x80ffffff));
        mFont->drawString(WIDTH / 2, 350, buf, IF_ALIGN_HORIZ_CENTER, 0.8f, Color(0xccffffff));
        if (mIMGUI.button(WIDTH / 2 - mButtonStyle->getWidth() / 2, 450, IMGUI_AI_COUNTER, mButtonStyle, 0, 0, "Quit"))
            App::getInstance()->setAlive(false);
    }

    mIMGUI.end();
    g->endRender();

    driver->endScene();
}
